;;;
;;; elevation profile socket client
;;;
;;;   Copyright (c) 2020 Jens Thiele <karme@karme.de>
;;;   
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;   
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;  
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;  
;;;   3. Neither the name of the authors nor the names of its contributors
;;;      may be used to endorse or promote products derived from this
;;;      software without specific prior written permission.
;;;  
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;  
(define-module elevation-profile-socket-client-v2
  (use util.list)
  (use gauche.sequence)
  (use gauche.time)
  (use gauche.net)
  (export elevation-profile-socket-open
          elevation-profile-socket-close
          polyline->3d
          upsample-polyline->4d
          sample-polyline->4d))

(select-module elevation-profile-socket-client-v2)

(define (retry f count . args)
  (let-optionals* args ((onerrorf (lambda(e)
                                    (with-output-to-port (current-error-port)
                                      (lambda()
                                        (print "error: " (ref e 'message e))
                                        (sys-sleep 1))))))
    (let loop ((count count))
      (guard (e
              [else
               (onerrorf e)
               (when (not (> count 0))
                 (raise e))
               (loop (- count 1))])
             (f)))))

(define (elevation-profile-socket-open . args)
  (let-optionals* args ((socket-spec '(inet "localhost" 2223)))
    (rlet1 this (make-hash-table) ;; todo: could also just use gauche's object system
           (letrec ((connect-2 (lambda() (apply make-client-socket socket-spec)))
                    (assert-protocol (lambda l (let1 r (remote-eval '(protocol-version))
                                                 (when (not (and (list? r)
                                                                 (equal? l r)))
                                                   (error "protocol mismatch" r)))))
                    (connect (lambda()
                               (hash-table-put! this 'socket (retry connect-2 5))
                               (assert-protocol 1 0)))
                    (reconnect (lambda()
                                 (when (hash-table-exists? this 'socket)
                                   (socket-close (hash-table-get this 'socket)))
                                 (connect)))
                    (remote-eval-2 (lambda(sexp)
                                     (let1 s (hash-table-get this 'socket)
                                       (with-ports (socket-input-port s) (socket-output-port s) #f
                                                   (lambda()
                                                     (write sexp)
                                                     (newline)
                                                     (flush)
                                                     (let1 r (read) ;; todo: security!
                                                       (when (eof-object? r)
                                                         (error r))
                                                       r))))))
                    (remote-eval (lambda(sexp)
                                   (let1 r (retry (cut remote-eval-2 sexp)
                                                  5
                                                  (lambda(e)
                                                    (with-output-to-port (current-error-port)
                                                      (lambda()
                                                        (print "error: " (ref e 'message e) " => reconnect")
                                                        (reconnect)))))
                                     (when (and (list? r)
                                                (eq? (car r) 'error))
                                       (error r))
                                     r))))
             (connect)
             (hash-table-put! this
                              'polyline->3d
                              (lambda(pl) (remote-eval (list 'polyline->3d pl))))
             (hash-table-put! this
                              'upsample-polyline->4d
                              (lambda(pl dist) (remote-eval (list 'upsample-polyline->4d 'wgs84 pl dist))))
             (hash-table-put! this
                              'sample-polyline->4d
                              (lambda(pl count) (remote-eval (list 'sample-polyline->4d 'wgs84 pl count))))))))

(define (polyline->3d this pl)
  ((hash-table-get this 'polyline->3d) pl))

(define (upsample-polyline->4d this pl upsample . args)
  ((hash-table-get this 'upsample-polyline->4d) pl upsample))

(define (sample-polyline->4d this pl samples . args)
  ((hash-table-get this 'sample-polyline->4d) pl samples))

(define (elevation-profile-socket-close this)
  (socket-close (hash-table-get this 'socket))
  (hash-table-delete! this 'socket))
